/*****************************************************************************
*
*  dwOpenClient.cpp - MQTT lightweight Client Support
*  Copyright (c) 2014, 2015 ILS Technology
*
*  Redistribution and use in source and binary forms, with or without
*  modification, are permitted provided that the following conditions
*  are met:
*
*    Redistributions of source code must retain the above copyright
*    notice, this list of conditions and the following disclaimer.
*
*    Redistributions in binary form must reproduce the above copyright
*    notice, this list of conditions and the following disclaimer in the
*    documentation and/or other materials provided with the
*    distribution.
*
*    Neither the name of ILS Technology nor Telit nor the names of its
*    contributors may be used tWSo endorse or promote products derived
*    from this software without specific prior written permission.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
*  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
*  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
*  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
*  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
*  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
*  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
*  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
*  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
*****************************************************************************/
#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <termios.h>
#include <pthread.h>
#include <cctype>

#include "dwOpenClient.h"
#include "dwOpenReceiver.h"

dwOpenClient::dwOpenClient(const char *Server, int Port )
{
  dwOpenReceiver::iConnected = 0;
  dwOpenReceiver::iSession = 0;
  dwOpenReceiver::iAuthorized = 0;
  dwOpenReceiver::iActive = 1;
  dwOpenReceiver::iHandle = 0;
  dwOpenReceiver::iOKReceived = 1;
}


bool dwOpenClient::dwConnect( const char *ThingKey, const char *ThingName, const char *AppToken)
{
   unsigned char d[9] = {0x00,0x06,'M','Q','I','s','d','p',MQTTPROTOCOLVERSION};
	
    int length;
	
	dwOpenReceiver::iConnected = 0;
	int mlength = 0; 
	
	printf("Attempting to connect...\n");
	
	memset(mbuffer, 0, sizeof(mbuffer));
	
	msgId = 0;
    nextMsgId = msgId + 1;

    // Leave room in the buffer for header and variable length field
    length = 5;

	memcpy( &workBuffer[5], d, sizeof(d) );
	
    length = 5 + sizeof(d);
	
	//ID, User, & Password being provided
    workBuffer[length++] = 0x02 | 0x80 | (0x80>>1);

	workBuffer[length++] = ((MQTT_KEEPALIVE) >> 8);
	workBuffer[length++] = ((MQTT_KEEPALIVE) & 0xFF);
	
	length = this->writeString(ThingKey,workBuffer,length);

	length = writeString(ThingName,workBuffer,length);
	length = writeString(AppToken,workBuffer,length);

    formatMsg(MQTTCONNECT,workBuffer,length-5,&mlength,mbuffer);	
	
	printf("Xmit ---> %d bytes\n", mlength );
	dwOpenReceiver::hexdump(mbuffer, mlength);

	// Send MQTT Request to Cloud Server...
	write(fd, mbuffer, mlength);
	
	printf("Awaiting mqttConnect Response...\n");
	while( dwOpenReceiver::iConnected == 0 ) sleep( 1 );

	if ( dwOpenReceiver::iConnected > 0 )
		return( true );
	else
		return( false);
}


void dwOpenClient::dwDisconnect( void )
{
    int i;
	int mlength = 0;	

    printf("in disconnect ...\n");

    mlength = 2;

    mbuffer[0] = MQTTDISCONNECT;
    mbuffer[1] = 0;

	// Send MQTT Request to Cloud Server...
	printf("Sending Disconnect to Cloud...\n");	
	write(fd, mbuffer, mlength);
}

bool dwOpenClient::dwSubscribe( const char *topic)
{	
	int mlength = 0; 
    int length = 5;
	unsigned char workBuf[50];
	
	memset(mbuffer, 0, sizeof(mbuffer));
	
	memset( workBuf, 0, sizeof(workBuf));

    nextMsgId++;
    if (nextMsgId == 0) nextMsgId = 1;
   
    workBuf[length++] = (nextMsgId >> 8);
    workBuf[length++] = (nextMsgId & 0xFF);
    length = writeString(topic, workBuf,length);
    workBuf[length++] = 0; // Only do QoS 0 subs

    formatMsg(MQTTSUBSCRIBE|MQTTQOS1,workBuf,length-5, &mlength, mbuffer);
	
	printf("Xmit ---> %d bytes\n", mlength );
	dwOpenReceiver::hexdump(mbuffer, mlength);

	// Subscribing to Topic 'reply'...
	write(fd, mbuffer, mlength);	

	return(0);	
}


bool dwOpenClient::dwUnsubscribe(const char* topic )
{
    int length = 5;
	int mlength = 0; 
	
    nextMsgId++;
    if (nextMsgId == 0) nextMsgId = 1;
    
    workBuffer[length++] = (nextMsgId >> 8);
    workBuffer[length++] = (nextMsgId & 0xFF);
    length = writeString(topic, workBuffer,length);

    formatMsg(MQTTUNSUBSCRIBE|MQTTQOS1,workBuffer,length-5, &mlength, mbuffer);

    return true;
}

bool dwOpenClient::dwOpenATCommandPort( const char *szComPort )
{
    printf("Opening %s\n", szComPort);

    fd = open(szComPort, O_RDWR | O_NOCTTY | O_NDELAY);

    if (fd == -1) {
        perror(szComPort);
        return 1;
    }

    if (tcgetattr(fd, &serial) < 0) {
        perror("Getting configuration");
        return 1;
    }

    // Set up Serial Configuration
    serial.c_iflag = 0;
    serial.c_oflag = 0;
    serial.c_lflag = 0;
    serial.c_cflag = 0;

    serial.c_cc[VMIN] = 0;
    serial.c_cc[VTIME] = 0;

    serial.c_cflag = B115200 | CS8 | CREAD;

    tcsetattr(fd, TCSANOW, &serial); // Apply configuration
	
	tcflush( fd, TCIOFLUSH );  
	
	pthread_create( &thread, NULL, dwOpenReceiver::SerialPortReceiver, (void *) &fd );

	return(0);
}

bool dwOpenClient::dwCloseATCommandPort()
{
	close(fd);
	
	return(0);
}

bool dwOpenClient::dwModuleConfig( const char *OpenURL, const char *APN)
{
	// Disable Flow Control
	ATCommandWrite( fd, "AT&K=0" );
	sleep( 1 );

	ATCommandWrite( fd, "ATE1" );
	sleep( 1 );
	
	ATCommandWrite( fd, "AT+CGSN" );
	sleep( 1 );

#if(0)	
	ATCommandWrite( fd, "AT#CIMI" );
	sleep( 1 );

	ATCommandWrite( fd, "AT+CCID" );
	sleep( 1 );

	// Should be 3
	ATCommandWrite( fd, "AT#BND=3" );
	sleep( 2 );

	// Should be 2 (USA)
	ATCommandWrite( fd, "AT#AUTOBND=2" );
	sleep( 2 );

	// Should be 3 (USA = 3, EUROPE = 0)
	ATCommandWrite( fd, "AT#BND?" );
	sleep( 1 );

	// Should be 2
	ATCommandWrite( fd, "AT#AUTOBND?" );
	sleep( 1 );

	ATCommandWrite( fd, "AT+CGMI" );
	sleep( 1 );

	ATCommandWrite( fd, "AT+CGMM" );
	sleep( 1 );

	ATCommandWrite( fd, "AT+CGMR" );
	sleep( 1 );

	ATCommandWrite( fd, "AT+COPS?" );
	sleep( 1 );

#endif
	
	// ======================================================================================================
	// == Connect to Cellular Network...
	// ======================================================================================================

	sprintf( (char *) mbuffer, "AT+CGDCONT=1,\"IP\",\"%s\"", APN );
	ATCommandWrite( fd, (const char *) mbuffer );
	sleep( 1 );

	dwOpenReceiver::iOKReceived = 0;
	ATCommandWrite( fd, "AT#SGACT=1,0" );
	while( dwOpenReceiver::iOKReceived == 0 ) sleep( 1 );
	sleep( 2 );

	ATCommandWrite( fd, "AT#SGACT=1,1" );
	while( dwOpenReceiver::iConnected == 0 ) sleep( 1 );

	// Display Connected Status...
    printf("PDP Context is Established...  Waiting 5s!\n");
	sleep( 5 );

	// Connect to the dwOpen Server...
    printf("Opening Socket Connection to m2mAIR Cloud Server...\n");
	sprintf( (char *) mbuffer, "AT#SD=1,0,1883,\"%s\"", OpenURL );
	ATCommandWrite( fd, (const char *) mbuffer );
	while( dwOpenReceiver::iSession == 0 ) sleep( 1 );

	sleep( 2 );
		
	return( 0 );
}


bool dwOpenClient::dwLogPublish( const char *ThingKey, const char *Message )
{
	int mlength = 0; 
		
	jsonCreateLogPublish( ThingKey, Message, jsonMsg );
	
	memset(mbuffer, 0, sizeof(mbuffer));
	
	mqttPublish("api", jsonMsg, strlen(jsonMsg), false, &mlength, mbuffer);
	
	printf("Xmit ---> %d bytes\n", mlength );	
	dwOpenReceiver::hexdump(mbuffer, mlength);

	// Send MQTT Request to Cloud Server...
	write(fd, mbuffer, mlength);	

	return(0);
}

bool dwOpenClient::dwLocationPublish( const char *ThingKey, float Latitude, float Longitude, float Altitude, float Speed, float Heading )
{
	int mlength = 0; 
	
	jsonCreateLocationPublish( ThingKey, Latitude, Longitude, Altitude, Speed, Heading, jsonMsg );
	
	mlength = 0; memset(mbuffer, 0, sizeof(mbuffer));
	
	mqttPublish("api", jsonMsg, strlen(jsonMsg), false, &mlength, mbuffer);
	
	printf("Xmit ---> %d bytes\n", mlength );	
	dwOpenReceiver::hexdump(mbuffer, mlength);

	// Send MQTT Request to Cloud Server...
	write(fd, mbuffer, mlength);
	
	return(0);	
}

bool dwOpenClient::dwAttributePublish( const char *ThingKey, const char *Key, const char *Value )
{
	int mlength = 0;
	
	jsonCreateSetAttribute( ThingKey, Key, Value, jsonMsg );
	
	memset(mbuffer, 0, sizeof(mbuffer));
	
	mqttPublish("api", jsonMsg, strlen(jsonMsg), false, &mlength, mbuffer);
	
	printf("Xmit ---> %d bytes\n", mlength );
	dwOpenReceiver::hexdump(mbuffer, mlength);

	write(fd, mbuffer, mlength);
	
	return(0);		
}
	
bool dwOpenClient::dwPropertyPublish( const char *ThingKey, const char *Key, const char *Value )
{
	int mlength = 0;
	
	jsonCreatePropertyPublish( ThingKey, Key, Value, jsonMsg );
	
	memset(mbuffer, 0, sizeof(mbuffer));
	
	mqttPublish("api", jsonMsg, strlen(jsonMsg), false, &mlength, mbuffer);
	
	printf("Xmit ---> %d bytes\n", mlength );
	dwOpenReceiver::hexdump(mbuffer, mlength);

	write(fd, mbuffer, mlength);
	
	return(0);		
}
	
void dwOpenClient::dwPublish(const char* topic, const char * payload )
{
	int plength = strlen(payload); 

	int mlength = 0; 

	memset(mbuffer, 0, sizeof(mbuffer));
	
	mqttPublish(topic, payload, plength, false, &mlength, mbuffer);	
	
	printf("Xmit ---> %d bytes\n", mlength );	
	dwOpenReceiver::hexdump(mbuffer, mlength);

	// Send MQTT Request to Cloud Server...
	write(fd, mbuffer, mlength);		
}

int dwOpenClient::writeString( const char *string, unsigned char *buf, int pos)
{
    const char *idp = string;
    int i = 0;

    pos += 2;
    while(*idp)
    {
        buf[pos++] = *idp++;
        i++;
    }
    buf[pos-i-2] = (i >> 8);
    buf[pos-i-1] = (i & 0xFF);

    return pos;
}

void dwOpenClient::jsonCreateLogPublish( const char *thingKey, const char *msgText, char *jsonText )
{
    // Create POST Request/Data...
    msgId++;
    sprintf(jsonText, "{ \"%05d\": { \"command\": \"log.publish\", \"params\": {\"msg\": \"%s\", \"thingKey\": \"%s\" }}}", msgId, msgText, thingKey );

    return;
}

void dwOpenClient::jsonCreatePropertyPublish( const char *thingKey, const char *propKey, const char *propValue, char *jsonText )
{
    // Create POST Request/Data...
    msgId++;
    sprintf(jsonText, "{ \"%05d\": { \"command\": \"property.publish\", \"params\": {\"thingKey\": \"%s\",\"key\": \"%s\",\"value\": %s }}}", msgId, thingKey, propKey, propValue );

    return;
}

void dwOpenClient::jsonCreateAlarmPublish( const char *thingKey, const char *alarmKey, int alarmState, const char *alarmMsg, char *jsonText )
{
    // Create POST Request/Data...
    msgId++;
    sprintf(jsonText, "{ \"%05d\": { \"command\": \"alarm.publish\", \"params\": {\"thingKey\": \"%s\", \"key\": \"%s\", \"msg\": \"%s\", \"state\": %d}}}", msgId, thingKey, alarmKey, alarmMsg, alarmState );

    return;
}

void dwOpenClient::jsonCreateLocationPublish( const char *thingKey, float locLatitude, float locLongitude, float locAltitude, float locSpeed, float locHeading, char *jsonText )
{
    // Create POST Request/Data...
    msgId++;
    sprintf(jsonText, "{ \"%05d\": { \"command\": \"location.publish\", \"params\": {\"thingKey\": \"%s\", \"lat\": \"%f\", \"lng\": %f, \"altitude\": %f, \"speed\": %f, \"alti\": %f}}}", msgId, thingKey, locLatitude, locLongitude, locAltitude, locSpeed, locHeading );

    return;
}

void dwOpenClient::jsonCreateSetAttribute( const char *thingKey, const char *attribKey, const char *attribValue, char *jsonText )
{
    // Create POST Request/Data...
    msgId++;
    sprintf(jsonText, "{ \"%05d\": { \"command\": \"thing.attr.set\", \"params\": {\"thingKey\": \"%s\", \"key\": \"%s\", \"value\": \"%s\"}}}", msgId, thingKey, attribKey, attribValue );

    return;
}

void dwOpenClient::jsonCreateGetAttribute( const char *thingKey, const char *attribKey, char *jsonText )
{
    // Create POST Request/Data...
    msgId++;
    sprintf(jsonText, "{ \"%05d\": { \"command\": \"thing.attr.get\", \"params\": {\"thingKey\": \"%s\", \"key\": \"%s\"}}}", msgId, thingKey, attribKey );

    return;
}


int dwOpenClient::ATCommandWrite( int fd, const char* szText )
{
	int iLength = strlen( szText );
	char szBuffer[512];
    int dwBytesWritten;

	printf( "Xmit ---> %s\n", szText );

	memset( szBuffer, 0, sizeof(szBuffer) );
	memcpy( szBuffer, szText, iLength );
	szBuffer[iLength] = '\r';

    dwBytesWritten = write(fd, szBuffer, strlen(szBuffer));
    if (dwBytesWritten < 0) {
        perror("Write");
        return -1;
    }
    else {
        printf("Sent %d characters\n", dwBytesWritten);
    }	

    return 0;
}


bool dwOpenClient::mqttPing(int conn_handle)
{
    int retval = 0;
    int i, len;
    int llen= 128;
    unsigned char type;
    int t1, t2;
    unsigned char topic[64];
    unsigned char payload[512];

    printf("in ping... \n");

    workBuffer[0] = MQTTPINGREQ;
    workBuffer[1] = 0;

    write( fd, workBuffer, 2 );

    printf("Client: MQTTPINGREQ send() is OK.\n");

    return true;
}


bool dwOpenClient::formatMsg(short header, unsigned char* buf, int length, int *mlength, unsigned char* mbuffer )
{
    int   i;

    short lenBuf[4];
    short llen = 0;
    short digit;
    short pos = 0;
    short rc;
    short len = length;

    do
    {
        digit = len % 128;
        len = len / 128;
        if (len > 0)
        {
            digit |= 0x80;
        }
        lenBuf[pos++] = digit;
        llen++;
    } while(len>0);

    buf[4-llen] = header;
    for(i=0; i<llen; i++)
    {
        buf[5-llen+i] = lenBuf[i];
    }

    *mlength = length+1+llen;

    memcpy(mbuffer, buf+(4-llen), *mlength);

    return true;
}


bool dwOpenClient::mqttPublish(const char* topic, const char* payload, unsigned int plength, bool retained, int *mlength, unsigned char *mbuffer)
{
    int i;
    int length = 5;
    short header = MQTTPUBLISH;

    // Leave room in the buffer for header and variable length field
    //printf("in publish ... %s\n",topic);
    //printf("in publish ... %s\n",payload);

    length = writeString(topic,workBuffer,length);

    for (i=0; i<plength; i++)
    {
        workBuffer[length++] = payload[i];
    }

    if (retained)
    {
        header |= 1;
    }

    formatMsg(header,workBuffer,length-5,mlength,mbuffer);

    return true;
}
