/*****************************************************************************
*
*  dwOpenClient.h - MQTT lightweight Client Support
*  Copyright (c) 2014, 2015 ILS Technology
*
*  Redistribution and use in source and binary forms, with or without
*  modification, are permitted provided that the following conditions
*  are met:
*
*    Redistributions of source code must retain the above copyright
*    notice, this list of conditions and the following disclaimer.
*
*    Redistributions in binary form must reproduce the above copyright
*    notice, this list of conditions and the following disclaimer in the
*    documentation and/or other materials provided with the
*    distribution.
*
*    Neither the name of ILS Technology nor Telit nor the names of its
*    contributors may be used tWSo endorse or promote products derived
*    from this software without specific prior written permission.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
*  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
*  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
*  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
*  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
*  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
*  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
*  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
*  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
*****************************************************************************/

#ifndef dwOpenClient_h
#define dwOpenClient_h

// MQTT_MAX_PACKET_SIZE : Maximum packet size
#define MQTT_MAX_PACKET_SIZE 1024

// MQTT_KEEPALIVE : keepAlive interval in Seconds
#define MQTT_KEEPALIVE 15

#define MQTTPROTOCOLVERSION 3
#define MQTTCONNECT 1 << 4 // Client request to connect to Server
#define MQTTCONNACK 2 << 4 // Connect Acknowledgment
#define MQTTPUBLISH 3 << 4 // Publish message
#define MQTTPUBACK 4 << 4 // Publish Acknowledgment
#define MQTTPUBREC 5 << 4 // Publish Received (assured delivery part 1)
#define MQTTPUBREL 6 << 4 // Publish Release (assured delivery part 2)
#define MQTTPUBCOMP 7 << 4 // Publish Complete (assured delivery part 3)
#define MQTTSUBSCRIBE 8 << 4 // Client Subscribe request
#define MQTTSUBACK 9 << 4 // Subscribe Acknowledgment
#define MQTTUNSUBSCRIBE 10 << 4 // Client Unsubscribe request
#define MQTTUNSUBACK 11 << 4 // Unsubscribe Acknowledgment
#define MQTTPINGREQ 12 << 4 // PING Request
#define MQTTPINGRESP 13 << 4 // PING Response
#define MQTTDISCONNECT 14 << 4 // Client is Disconnecting
#define MQTTReserved 15 << 4 // Reserved

#define MQTTQOS0 (0 << 1)
#define MQTTQOS1 (1 << 1)
#define MQTTQOS2 (2 << 1)

#define MQTT_REPLY		"reply"
#define MQTT_NOTIFY 	"notify"

#define true 1
#define false 0

class dwOpenClient 
{
	private:
	
		int fd;
		pthread_t thread;

		int msgId;
		int nextMsgId;

		struct termios serial;
		
		unsigned char mbuffer[MQTT_MAX_PACKET_SIZE];
		unsigned char workBuffer[MQTT_MAX_PACKET_SIZE];

		char jsonMsg[512];	
		
		int writeString( const char *, unsigned char *, int );
		bool mqttPing(int);
		bool formatMsg(short, unsigned char *, int, int *, unsigned char *);

		bool mqttPublish(const char *, const char *, unsigned int, bool, int *, unsigned char *);

		void jsonCreateLogPublish( const char *, const char *, char * );
		void jsonCreatePropertyPublish( const char *, const char *, const char *, char * );
		void jsonCreateAlarmPublish( const char *, const char *, int , const char *, char * );
		void jsonCreateLocationPublish( const char *, float, float, float, float, float, char * );
		void jsonCreateSetAttribute( const char *, const char *, const char *, char * );
		void jsonCreateGetAttribute( const char *, const char *, char * );
	 
		int ATCommandWrite( int, const char * );	
		
	public:
	
		dwOpenClient(const char *, int );
		
		bool dwOpenATCommandPort( const char * );
		bool dwCloseATCommandPort( void );
		bool dwModuleConfig( const char *, const char *);	

		bool dwConnect( const char*, const char*, const char* );
		void dwDisconnect( void );	

		bool dwSubscribe( const char * );
		bool dwUnsubscribe(const char* );	
		bool dwLogPublish( const char *, const char * );
		bool dwLocationPublish( const char *, float, float, float, float, float );
		bool dwAttributePublish( const char *, const char *, const char * );
		bool dwPropertyPublish( const char *, const char *, const char * );
		
		void dwPublish(const char* , const char * );

};




#endif
