/*****************************************************************************
*
*  dwOpenReceiver.h - Receiver handler for MQTT lightweight Client Support
*  Copyright (c) 2014, 2015 ILS Technology
*
*  Redistribution and use in source and binary forms, with or without
*  modification, are permitted provided that the following conditions
*  are met:
*
*    Redistributions of source code must retain the above copyright
*    notice, this list of conditions and the following disclaimer.
*
*    Redistributions in binary form must reproduce the above copyright
*    notice, this list of conditions and the following disclaimer in the
*    documentation and/or other materials provided with the
*    distribution.
*
*    Neither the name of ILS Technology nor Telit nor the names of its
*    contributors may be used tWSo endorse or promote products derived
*    from this software without specific prior written permission.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
*  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
*  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
*  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
*  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
*  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
*  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
*  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
*  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
*****************************************************************************/

extern dwOpenClient dwClient;

namespace dwOpenReceiver
{
	int  iConnected;
	int  iSession;
	int  iAuthorized;
	int  iActive;
	int  iHandle;
	int  iOKReceived;
	
	int fd;
	
	unsigned char recvBuffer[MQTT_MAX_PACKET_SIZE];
	
	void hexdump(void *mem, unsigned int len);
	void *SerialPortReceiver( void * );
	void ResponseHandler( int );
		
}

void *dwOpenReceiver::SerialPortReceiver( void *fileHandle )
{
	dwOpenReceiver::fd = *( (int *) fileHandle);
	
	memset( dwOpenReceiver::recvBuffer, 0, sizeof(dwOpenReceiver::recvBuffer));	
	
	while(1)
	{			
		int rcount = read( dwOpenReceiver::fd, dwOpenReceiver::recvBuffer, sizeof(dwOpenReceiver::recvBuffer));		
		if (rcount < 0) {
			perror("Serial Port Read Error");
		}
		else if (rcount == 0)
		{
			sleep(1);			
		}
		else 
		{			
			dwOpenReceiver::recvBuffer[rcount] = '\0';	
			printf("Recv ---> %d bytes\n", rcount );			
			dwOpenReceiver::hexdump(dwOpenReceiver::recvBuffer, rcount);			
			dwOpenReceiver::ResponseHandler(rcount);
			
			memset( dwOpenReceiver::recvBuffer, 0, sizeof(dwOpenReceiver::recvBuffer));

		}
	}
}

//TR50 JSON Sample Request/Response
#define MAILBOX_IN_PROGRESS_ACK  "{\"999-999\": { \"command\": \"mailbox.check\", \"params\": { \"autoComplete\": true, \"limit\": 0 }}}"
#define MAILBOX_REQ_RECEIVED     "{\"0000123\": { \"command\": \"mailbox.update\", \"params\": {\"id\": \"%s\", \"msg\": \"Executing...\"}}}"
#define MAILBOX_REQ_ACK          "{\"0000124\": { \"command\": \"mailbox.ack\", \"params\": {\"id\": \"%s\", \"errorCode\": \"0\", \"errorMessage\": \"OK\"}}}"

void ExtractKeyValue( char *haystack, const char *Key, char *ResultValue )
{
	char *KeyPtr;
	char *KeyValue;
	char *KeyValueEnd;
	int   KeyValueLen;

	char needleTmp[40];
		
	strcpy( needleTmp, "\"");
	strcat( needleTmp, Key);
	strcat( needleTmp, "\":\"");

	KeyPtr = strstr(haystack, needleTmp );
		
	if( KeyPtr != NULL) 
	{	  
		KeyValue = KeyPtr + strlen(needleTmp);
		KeyValueEnd = KeyValue;
		
		while( *KeyValueEnd != '\"')
		{
			KeyValueEnd++;
		}
			
		KeyValueLen = KeyValueEnd - KeyValue;
		strncpy( ResultValue, KeyValue, KeyValueLen );
		*(ResultValue + KeyValueLen ) = '\0';
	}
	else
	{
		strcpy( ResultValue, "" );
	}

}

void dwOpenReceiver::ResponseHandler( int BytesRead )
{
	unsigned char type;
		
	type = dwOpenReceiver::recvBuffer[0] & 0xF0;
	printf("MQTT Message Type = 0x%02X\n", type);
	
	// Check Response Status (MQTTCONNACK = 20 02 00 00)
	if(BytesRead == 4 && type == MQTTCONNACK)
	{
		printf("MQTTCONNACK Received...\n");

		if(dwOpenReceiver::recvBuffer[3] == 0)
		{
			printf("MQTT Connect Successful...\n");
			dwOpenReceiver::iConnected = 1;
		}
		else
		{
			printf("MQTT Connect Failed...\n");
			dwOpenReceiver::iConnected = -1;
		}
	}
	else if(type == MQTTPUBLISH)
	{
		//printf("Type: MQTTPUBLISH... \n");
		char jsonText[1024];   
		char msgID[50];			
		char LEDStatus[10]; 		
		
		int payloadLen = (dwOpenReceiver::recvBuffer[2]<<8)+dwOpenReceiver::recvBuffer[1];
		int topicLenOffset = dwOpenReceiver::recvBuffer[3] ? 3 : 4;
		int topicLen = (int) dwOpenReceiver::recvBuffer[topicLenOffset];
		
		char topic[topicLen+1];
		
		memcpy( topic, &dwOpenReceiver::recvBuffer[topicLenOffset+1], topicLen );
		topic[topicLen] = '\0';
		
		printf("Topic Name: %s\n", topic );
		
		char *payload = (char *) &dwOpenReceiver::recvBuffer[topicLen+4];
		
		if(strstr(topic, MQTT_NOTIFY) != NULL) 
		{  
			printf("Notification - Mailbox Activity... Check Mailbox\n");
			dwClient.dwPublish("api", MAILBOX_IN_PROGRESS_ACK );    
		}
		else if( (strstr(topic, MQTT_REPLY) != NULL) )
		{
			if ((strstr((char *)payload, "999-999") != NULL)) 
			{     
				printf("Reply - Mailbox Check Response... Mailbox Request Received\n");   
				ExtractKeyValue( payload, "id", msgID ); 

				sprintf( jsonText, MAILBOX_REQ_RECEIVED, msgID );  
				dwClient.dwPublish("api", jsonText );

				//Set LED accordingly                 
				ExtractKeyValue( (char *) payload, "greenled", LEDStatus );                
				if ( strlen(LEDStatus) ) printf("Simulated Green LED Status: %s\n", LEDStatus );

				ExtractKeyValue( (char *) payload, "redled", LEDStatus );                
				if ( strlen(LEDStatus) ) printf("Simulated Red LED Status: %s\n", LEDStatus );
				  
				//Serial.println("Processing Complete - Send Mailbox Acknowledge...");
				sprintf( jsonText, MAILBOX_REQ_ACK, msgID );   
				dwClient.dwPublish( "api", jsonText );  
			} 

		}		
	}	
	else if(type == MQTTPINGREQ)
	{
		printf("Type: MQTTPINGREQ... \n");

		dwOpenReceiver::recvBuffer[0] = MQTTPINGRESP;
		dwOpenReceiver::recvBuffer[1] = 0;

		//SerialPortWriteBinary( hSerial, 2, buffer );
		write(dwOpenReceiver::fd, dwOpenReceiver::recvBuffer, 2);

		printf("Client: MQTTPINGRESP send() is OK.\n");
	}
	else if(type == MQTTPINGRESP && BytesRead > 2)
	{
		printf(">>>> COMPOUND Message Received, Processing In-Line...\n");

		printf("Type: MQTTPINGRESP... \n");

		//pingOutstanding = false;
	}
	else if(type == MQTTPINGRESP)
	{
		printf("Type: MQTTPINGRESP... \n");

		//pingOutstanding = false;
	}

	
	// Check for OK Messages...
	if( dwOpenReceiver::iOKReceived == 0 && strstr( (char *) dwOpenReceiver::recvBuffer, "OK\r\n" ) > 0 )
	{
		printf("OK Message Received...\n");
		dwOpenReceiver::iOKReceived = 1;
	}

	// Check for SGACT Messages...
	if( strstr( (char *) dwOpenReceiver::recvBuffer, "SGACT:" ) > 0 )
	{
		printf("#SGACT Message Received, PDP Context Activated...\n");
		dwOpenReceiver::iConnected = 1;
	}

	// Check for CONNECT Messages...
	if( strstr( (char *) dwOpenReceiver::recvBuffer, "CONNECT\r\n" ) > 0 )
	{
		printf("CONNECT Message Received...\n");
		dwOpenReceiver::iSession = 1;
	}

	// Check for NO CARRIER Messages...
	if( strstr( (char *) dwOpenReceiver::recvBuffer, "NO CARRIER\r\n" ) > 0 )
	{
		printf("NO CARRIER Message Received...\n");
		dwOpenReceiver::iSession = 0;
	}
	
}

void dwOpenReceiver::hexdump(void *mem, unsigned int len)
{
	#define HEXDUMP_COLS 16
	
	unsigned int i, j;

	for(i = 0; i < len + ((len % HEXDUMP_COLS) ? (HEXDUMP_COLS - len % HEXDUMP_COLS) : 0); i++)
	{
			/* print offset */
			if(i % HEXDUMP_COLS == 0)
			{
				printf("0x%06X: ", i);
			}

			/* print hex data */
			if(i < len)
			{
					printf("%02X ", 0xFF & ((char*)mem)[i]);
			}
			else /* end of block, just aligning for ASCII dump */
			{
					printf("   ");
			}

			/* print ASCII dump */
			if(i % HEXDUMP_COLS == (HEXDUMP_COLS - 1))
			{
					for(j = i - (HEXDUMP_COLS - 1); j <= i; j++)
					{
							if(j >= len) /* end of block, not really printing */
							{
									putchar(' ');
							}
							else if(std::isprint(((char*)mem)[j])) /* printable char */
							{
									putchar(0xFF & ((char*)mem)[j]);
							}
							else /* other char */
							{
									putchar('.');
							}
					}
					putchar('\n');
			}
	}
}



