/*****************************************************************************
*
*  dwRaspberryPi.c - Raspberry Pi deviceWISE Sample 'Thing' Program
*  Copyright (c) 2014, 2015 ILS Technology
*
*  Redistribution and use in source and binary forms, with or without
*  modification, are permitted provided that the following conditions
*  are met:
*
*    Redistributions of source code must retain the above copyright
*    notice, this list of conditions and the following disclaimer.
*
*    Redistributions in binary form must reproduce the above copyright
*    notice, this list of conditions and the following disclaimer in the
*    documentation and/or other materials provided with the
*    distribution.
*
*    Neither the name of ILS Technology nor Telit nor the names of its
*    contributors may be used tWSo endorse or promote products derived
*    from this software without specific prior written permission.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
*  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
*  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
*  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
*  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
*  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
*  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
*  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
*  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
*
*	To Compile:
*   g++ -lpthread dwRaspberryPi.c dwOpenClient.cpp -o RaspberryPi
*
*
*****************************************************************************/
#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <termios.h>
#include <pthread.h>

#include "dwOpenClient.h"

// Authentication/Registration Details  
#define DWOPEN_THING_NAME  	"RaspberryPi"		
#define DWOPEN_APPTOKEN    	"1kagVL7HGh6JAd9w"		//Application Token


#define DWOPEN_SERVER      	"api.devicewise.com" 	//deviceWISE server
#define DWOPEN_PORT        	1883                 	//MQTT Port

//#define DWOPEN_APN			"internetm2m.air.com"  //Specify APN for connecting to Cellular Carrier...
#define DWOPEN_APN			"a105.2way.net"  
	

//ILST Headquarters - GPS Detail...
#define ILST_Latitude      26.394919
#define ILST_Longitude     -80.112808
#define ILST_Altitude      3.0
#define ILST_Speed         0.0
#define ILST_Heading       0.0

dwOpenClient dwClient( (const char *) DWOPEN_SERVER, DWOPEN_PORT );

int main(int argc, char* argv[]) 
{
    if (argc == 1) {
        printf("Usage: %s [device]\n\n", argv[0]);
        return -1;
    }
	
	if ( dwClient.dwOpenATCommandPort( argv[1] ) )
	{
		printf("\nMake sure Serial device is plugged in!\n");
		return( -1 );
	}
	
	//Setup Module for TR50
	dwClient.dwModuleConfig(DWOPEN_SERVER, DWOPEN_APN);
	
	//MQTT Commands start here...

	printf("Connecting to the Cloud...\n");
	dwClient.dwConnect( DWOPEN_THING_NAME, DWOPEN_THING_NAME, DWOPEN_APPTOKEN);	
	sleep(10);
   	
	printf("Subscribing to Topic 'reply'...");
	dwClient.dwSubscribe(MQTT_REPLY);	
	sleep(10);
	
	printf("Subscribing to Topic 'notify'...");
	dwClient.dwSubscribe(MQTT_NOTIFY);
	sleep(10);	

	printf("mqttClientApp - Log Publish\n" );
	dwClient.dwLogPublish( DWOPEN_THING_NAME, "Raspberry Pi Telit Module Connected on MQTT (Native)...");	
	sleep(5);

	printf("mqttClientApp - Location Publish\n" );	
	dwClient.dwLocationPublish( DWOPEN_THING_NAME, ILST_Latitude, ILST_Longitude, ILST_Altitude, ILST_Speed, ILST_Heading ); 	
	sleep(5);


	printf("mqttClientApp - Set Attribute\n" );
	dwClient.dwAttributePublish( DWOPEN_THING_NAME, "ID", "A12345678" );  
	sleep(5);

	printf("mqttClientApp - Property Publish\n" );
	dwClient.dwPropertyPublish( DWOPEN_THING_NAME, "Temp", "70.1" );  
	sleep(5);

	printf("Going idle for a 90 secs\n");
	sleep(90);
	
	dwClient.dwDisconnect();	
	sleep(5);
	
	printf("\nBye!\n");
	
	dwClient.dwCloseATCommandPort();
    
}
